use crate::{builder::ServerBuilder, config::ServerConfig, responder::Responder};
use http1::Builder;
use http_body_util::Full;
use hyper::{body::Incoming, server::conn::http1, service::service_fn, Request, Response};
use hyper_util::rt::{TokioIo, TokioTimer};
use log::error;
use std::{convert::Infallible, future::Future, net::SocketAddr, sync::Arc};
use tokio::{net::TcpListener, spawn};
use tokio_util::bytes::Bytes;

pub struct Server {
    pub config: Arc<ServerConfig>,
}

impl Server {
    pub fn builder() -> ServerBuilder {
        ServerBuilder {
            config: ServerConfig::default(),
        }
    }

    pub async fn run<F, Fut>(self, handler: F)
    where
        F: Fn(Request<Incoming>) -> Fut + Send + Sync + 'static,
        Fut: Future<Output = Result<Response<Full<Bytes>>, Infallible>> + Send,
    {
        let addr: SocketAddr = format!("{}:{}", self.config.ip, self.config.port)
            .parse()
            .expect("Invalid IP or port");
        let listener = TcpListener::bind(addr).await.unwrap();

        let handler = Arc::new(handler);
        loop {
            let (tcp, client_addr) = match listener.accept().await {
                Ok(conn) => conn,
                Err(error) => {
                    error!(
                        error = error.to_string().as_str();
                        "Failed to accept connection"
                    );
                    continue;
                }
            };
            let io = TokioIo::new(tcp);

            let config = Arc::clone(&self.config);
            let handler = Arc::clone(&handler);
            spawn(async move {
                if let Err(error) = Builder::new()
                    .timer(TokioTimer::new())
                    .serve_connection(
                        io,
                        service_fn(move |req| {
                            let config = Arc::clone(&config);
                            let handler = Arc::clone(&handler);

                            async move {
                                if !config.is_ip_authorized(&client_addr.ip())
                                    || !config.is_req_authorized(&req)
                                {
                                    if config.log_unauthorized {
                                        error!(tag = "ban",
                                            ip = client_addr.ip().to_string().as_str();
                                            "Unauthorized"
                                        );
                                    }

                                    Responder::unathorized()
                                } else {
                                    handler(req).await
                                }
                            }
                        }),
                    )
                    .await
                {
                    error!(error = error.to_string().as_str();
                        "Serving connection"
                    );
                }
            });
        }
    }
}
